import logging
import multiprocessing
from typing import Optional

from jsoncfg.value_mappers import require_string, require_integer, require_bool

logger = logging.getLogger(__name__)


class PeekFileConfigWorkerMixin:
    @property
    def celeryBrokerUrl(self) -> str:
        # for BROKER_URL
        default = 'amqp://guest:guest@localhost:5672//'
        with self._cfg as c:
            return c.celery.brokerUrl(default, require_string)

    @property
    def celeryResultUrl(self) -> str:
        # for CELERYD_CONCURRENCY
        default = 'redis://localhost:6379/0'
        with self._cfg as c:
            return c.celery.resultUrl(default, require_string)

    @property
    def celeryWorkerCount(self) -> str:
        # for CELERYD_CONCURRENCY

        # By default, we assume a single server setup.
        # So leave half the CPU threads for the database
        default = int(multiprocessing.cpu_count() / 2)
        with self._cfg as c:
            return c.celery.worker.workerCount(default, require_integer)

    @property
    def celeryTaskPrefetch(self) -> str:
        # for CELERYD_PREFETCH_MULTIPLIER

        default = 2

        with self._cfg as c:
            return c.celery.worker.taskPrefetch(default, require_integer)

    @property
    def celeryReplaceWorkerAfterTaskCount(self) -> Optional[int]:
        # for worker_max_tasks_per_child
        with self._cfg as c:
            try:
                return int(c.celery.worker.refreshAfterTaskCount(None))
            except TypeError:
                return None
            except ValueError:
                return None

    @property
    def celeryReplaceWorkerAfterMemUsage(self) -> Optional[int]:
        # for worker_max_memory_per_child
        with self._cfg as c:
            try:
                return int(c.celery.worker.refreshAfterMemUsage(None))
            except TypeError:
                return None
            except ValueError:
                return None

    @property
    def celeryConnectionPoolSize(self) -> int:
        with self._cfg as c:
            return c.celery.caller.connectionPoolSize(50, require_integer)

    @property
    def celeryConnectionRecycleTime(self) -> int:
        with self._cfg as c:
            return c.celery.caller.connectionRecycleTime(600, require_integer)

    @property
    def celeryConnectionPoolSize(self) -> int:
        with self._cfg as c:
            return c.celery.caller.connectionPoolSize(50, require_integer)

    @property
    def celeryPlPythonEnablePatch(self) -> int:
        with self._cfg as c:
            return c.celery.plpython.enable(False, require_bool)

    @property
    def celeryPlPythonWorkerCount(self) -> int:
        with self._cfg as c:
            return c.celery.plpython.workerCount(6, require_integer)



import logging
from datetime import datetime
from typing import List

import os
import pytz

from peek_platform.build_common.BuilderOsCmd import runDocBuild
from peek_platform.build_doc.DocBuilderABC import DocBuilderABC
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class DocBuilder(DocBuilderABC):

    def __init__(self, docProjectDir: str, platformService: str,
                 jsonCfg, loadedPlugins: List):
        DocBuilderABC.__init__(self, docProjectDir, platformService,
                               jsonCfg, loadedPlugins)

    @deferToThreadWrapWithLogger(logger, checkMainThread=False)
    def build(self) -> None:
        if not self._jsonCfg.docBuildPrepareEnabled:
            logger.info("%s SKIPPING, Doc build prepare is disabled in config",
                        self._platformService)
            return

        excludeRegexp = (
            r'.*__pycache__.*',
            r'.*[.]py$'
        )

        docLinkDir = os.path.join(self._docProjectDir, "doc_link")
        docConfDir = self._docProjectDir
        docDistDir = os.path.join(self._docProjectDir, "doc_dist")

        if not os.path.isdir(docLinkDir):
            os.mkdir(docLinkDir)

        self._dirSyncMap = list()

        pluginDetails = self._loadPluginConfigs()

        pluginDetails.sort(key=lambda item: item.pluginTitle)

        # --------------------
        # Prepare the table of contents link ins
        self._writePluginsToc(docLinkDir, pluginDetails)
        self._writePluginToc(docLinkDir, pluginDetails)

        # --------------------
        # Prepare the API document loads
        self._writePluginsApiConf(docLinkDir, pluginDetails)
        self._writePluginsApiList(docLinkDir, pluginDetails)

        # --------------------
        # Now sync the plugins documentation directory
        self._syncPluginFiles(docLinkDir,
                              pluginDetails,
                              excludeFilesRegex=excludeRegexp)

        self.fileSync.syncFiles()

        if self._jsonCfg.docSyncFilesForDebugEnabled:
            logger.info("%s starting frontend development file sync",
                        self._platformService)
            self.fileSync.startFileSyncWatcher()

        if self._jsonCfg.docBuildEnabled:
            logger.info("%s starting frontend web build", self._platformService)
            self._compileDocs(docLinkDir)

    def _syncFileHook(self, fileName: str, contents: bytes) -> bytes:
        return contents

    def _compileDocs(self, docLinkDir: str) -> None:
        """ Compile the docs

        this runs `build_html_docs.sh`, after checking if any files have changed.

        """
        startDate = datetime.now(pytz.utc)
        hashFileName = os.path.join(docLinkDir, ".lastHash")

        if not self._recompileRequiredCheck(docLinkDir, hashFileName):
            logger.info("%s Doc has not changed, recompile not required.",
                        self._platformService)
            return

        logger.info("%s Rebuilding docs for ", self._platformService)

        try:
            runDocBuild(self._docProjectDir)

        except Exception as e:
            if os.path.exists(hashFileName):
                os.remove(hashFileName)

            # Update the detail of the exception and raise it
            e.message = "%s sphinx docs failed to build." % self._platformService
            raise

        logger.info("%s sphinx doc rebuild completed in %s",
                    self._platformService, datetime.now(pytz.utc) - startDate)



'''
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
 *
'''

import logging
import os
from abc import ABCMeta

from jsoncfg.functions import save_config, ConfigWithWrapper

logger = logging.getLogger(__name__)


class PeekFileConfigABC(metaclass=ABCMeta):
    """
    This class creates a basic agent configuration
    """

    DEFAULT_FILE_CHMOD = 0o600
    DEFAULT_DIR_CHMOD = 0o700

    __instance = None

    def __new__(cls):
        if cls.__instance is not None:
            return cls.__instance

        self = super(PeekFileConfigABC, cls).__new__(cls)
        cls.__instance = self
        return self

    def __init__(self):
        '''
        Constructor
        '''
        from peek_platform import PeekPlatformConfig
        assert PeekPlatformConfig.componentName is not None

        self._homePath = os.path.join(
            os.path.expanduser('~'),
            '%s.home' % PeekPlatformConfig.componentName)

        if not os.path.isdir(self._homePath):
            assert (not os.path.exists(self._homePath))
            os.makedirs(self._homePath, self.DEFAULT_DIR_CHMOD)

        self._configFilePath = os.path.join(self._homePath, 'config.json')

        if not os.path.isfile(self._configFilePath):
            assert (not os.path.exists(self._configFilePath))
            with open(self._configFilePath, 'w') as fobj:
                fobj.write('{}')

        self._cfg = ConfigWithWrapper(self._configFilePath)

        self._hp = '%(' + self._homePath + ')s'

    def _save(self):
        save_config(self._configFilePath, self._cfg)

    def _chkDir(self, path):
        if not os.path.isdir(path):
            assert not os.path.exists(path)
            os.makedirs(path, self.DEFAULT_DIR_CHMOD)
        return path



import json
import logging
import os
import shutil
from collections import namedtuple
from textwrap import dedent
from typing import List, Callable, Optional, Dict

from jsoncfg.value_mappers import require_bool

from peek_platform.build_common.BuilderABC import BuilderABC
from peek_platform.build_frontend.FrontendFileSync import FrontendFileSync
from peek_platform.file_config.PeekFileConfigFrontendDirMixin import \
    PeekFileConfigFrontendDirMixin
from peek_platform.file_config.PeekFileConfigOsMixin import PeekFileConfigOsMixin
from peek_plugin_base.PluginPackageFileConfig import PluginPackageFileConfig

logger = logging.getLogger(__name__)

# Quiten the file watchdog
logging.getLogger("watchdog.observers.inotify_buffer").setLevel(logging.INFO)

PluginDetail = namedtuple("PluginDetail",
                          ["pluginRootDir",
                           "pluginName",
                           "pluginTitle",
                           "appDir",
                           "appModule",
                           "cfgDir",
                           "cfgModule",
                           "moduleDir",
                           "assetDir",
                           "rootModules",
                           "rootServices",
                           "rootComponents",
                           "icon",
                           "homeLinkText",
                           "showHomeLink",
                           "showInTitleBar",
                           "titleBarLeft",
                           "titleBarText",
                           "configLinkPath"])


class BuildTypeEnum:
    ELECTRON = "ELECTRON"
    WEB_DESKTOP = "WEB_DESKTOP"
    WEB_MOBILE = "WEB_MOBILE"
    WEB_ADMIN = "WEB_ADMIN"
    NATIVE_SCRIPT = "NATIVE_SCRIPT"


class FrontendBuilderABC(BuilderABC):
    """ Peek App Frontend Installer Mixin

    This class is used for the client and server.

    This class contains the logic for:
        * Linking in the frontend angular components to the frontend project
        * Compiling the frontend project

    :TODO: Use find/sort to generate a string of the files when this was last run.
        Only run it again if anything has changed.

    """

    _CFG_KEYS = {
        BuildTypeEnum.ELECTRON: ["desktop-electron", "desktop"],
        BuildTypeEnum.WEB_DESKTOP: ["desktop-web", "desktop"],
        BuildTypeEnum.WEB_MOBILE: ["mobile-web", "mobile"],
        BuildTypeEnum.NATIVE_SCRIPT: ["mobile-ns", "mobile"],
        BuildTypeEnum.WEB_ADMIN: ["admin"]
    }

    def __init__(self, frontendProjectDir: str, platformService: str,
                 buildType: BuildTypeEnum, jsonCfg,
                 loadedPlugins: List):
        assert platformService in ("peek-mobile", "peek-admin", "peek-desktop"), (
                "Unexpected service %s" % platformService)

        self._platformService = platformService
        self._buildType = buildType
        self._jsonCfg = jsonCfg
        self._frontendProjectDir = frontendProjectDir
        self._loadedPlugins = loadedPlugins

        if not isinstance(self._jsonCfg, PeekFileConfigFrontendDirMixin):
            raise Exception("The file config must inherit the"
                            " PeekFileConfigFrontendDirMixin")

        if not isinstance(self._jsonCfg, PeekFileConfigOsMixin):
            raise Exception("The file config must inherit the"
                            " PeekFileConfigOsMixin")

        if not os.path.isdir(frontendProjectDir):
            raise Exception("% doesn't exist" % frontendProjectDir)

        self.fileSync = FrontendFileSync(lambda f, c: self._syncFileHook(f, c))
        self._dirSyncMap = list()
        self._fileWatchdogObserver = None

    def _loadPluginConfigs(self) -> [PluginDetail]:
        pluginDetails = []

        for plugin in self._loadedPlugins:
            assert isinstance(plugin.packageCfg, PluginPackageFileConfig)
            pluginPackageConfig = plugin.packageCfg.config
            jsonCfgNode = None

            for configKey in self._CFG_KEYS[self._buildType]:
                if configKey in pluginPackageConfig:
                    jsonCfgNode = pluginPackageConfig[configKey]
                    break

            if not jsonCfgNode:
                logger.info("Skipping frontend build for %s,"
                            "missing config section for %s",
                            plugin.name, self._buildType)
                continue

            enabled = (jsonCfgNode.enableAngularFrontend(True, require_bool))

            if not enabled:
                continue

            appDir = jsonCfgNode.appDir(None)
            appModule = jsonCfgNode.appModule(None)

            cfgDir = jsonCfgNode.cfgDir(None)
            cfgModule = jsonCfgNode.cfgModule(None)

            moduleDir = jsonCfgNode.moduleDir(None)
            assetDir = jsonCfgNode.assetDir(None)

            showHomeLink = jsonCfgNode.showHomeLink(True)
            homeLinkText = jsonCfgNode.homeLinkText(plugin.title)
            showInTitleBar = jsonCfgNode.showInTitleBar(False)
            titleBarLeft = jsonCfgNode.titleBarLeft(False)
            titleBarText = jsonCfgNode.titleBarText(None)
            configLinkPath = jsonCfgNode.configLinkPath(None)

            def checkRootComponents(rootComponent):
                rootComponent["selector"] = rootComponent.get("selector")

            def checkThing(name, data):
                sub = (name, plugin.name)
                if data:
                    assert data["file"], "%s.file is missing for %s" % sub
                    assert data["class"], "%s.class is missing for %s" % sub

                if not data.get("persistent"):
                    data["persistent"] = False

                if not data.get("locatedInAppDir"):
                    data["locatedInAppDir"] = False

                # For services
                data["useClassFile"] = data.get("useClassFile")
                data["useClassClass"] = data.get("useClassClass")
                data["useExistingClass"] = data.get("useExistingClass")

            # Root Modules
            rootModules = jsonCfgNode.rootModules([])
            for rootModule in rootModules:
                checkThing("rootModules", rootModule)

            # Root Services
            rootServices = jsonCfgNode.rootServices([])
            for rootService in rootServices:
                checkThing("rootServices", rootService)

            # Root Components
            rootComponents = jsonCfgNode.rootComponents([])
            for rootComponent in rootComponents:
                checkRootComponents(rootComponent)

            icon = (jsonCfgNode.icon(None))

            pluginDetails.append(
                PluginDetail(pluginRootDir=plugin.rootDir,
                             pluginName=plugin.name,
                             pluginTitle=plugin.title,
                             appDir=appDir,
                             appModule=appModule,
                             cfgDir=cfgDir,
                             cfgModule=cfgModule,
                             moduleDir=moduleDir,
                             assetDir=assetDir,
                             rootModules=rootModules,
                             rootServices=rootServices,
                             rootComponents=rootComponents,
                             icon=icon,
                             homeLinkText=homeLinkText,
                             showHomeLink=showHomeLink,
                             showInTitleBar=showInTitleBar,
                             titleBarLeft=titleBarLeft,
                             titleBarText=titleBarText,
                             configLinkPath=configLinkPath)
            )

        pluginDetails.sort(key=lambda x: x.pluginName)
        return pluginDetails

    def _makeModuleOrServicePath(self, pluginDetail, modOrService):
        if modOrService["locatedInAppDir"]:
            return "%s" % pluginDetail.pluginName

        return "@peek/%s" % pluginDetail.pluginName

    def _writePluginHomeLinks(self, feAppDir: str,
                              pluginDetails: [PluginDetail]) -> None:
        """
        export const homeLinks = [
            {
                name: 'plugin_noop',
                title: "Noop",
                resourcePath: "/peek_plugin_noop",
                pluginIconPath: "/peek_plugin_noop/home_icon.png"
            }
        ];
        """

        links = []
        for pluginDetail in pluginDetails:
            if not (pluginDetail.appModule and pluginDetail.showHomeLink):
                continue

            links.append(dict(name=pluginDetail.pluginName,
                              title=pluginDetail.homeLinkText,
                              resourcePath="/%s" % pluginDetail.pluginName,
                              pluginIconPath=pluginDetail.icon))

        links.sort(key=lambda item: item["title"])

        contents = "// This file is auto generated, the git version is blank and .gitignored\n"
        contents += "export const homeLinks = %s;\n" % json.dumps(
            links, sort_keys=True, indent=4, separators=(', ', ': '))

        self._writeFileIfRequired(feAppDir, 'plugin-home-links.ts', contents)

    def _writePluginConfigLinks(self, feAppDir: str,
                                pluginDetails: [PluginDetail]) -> None:
        """
        export const configLinks = [
            {
                name: 'plugin_noop',
                title: "Noop",
                resourcePath: "/peek_plugin_noop_cfg",
                pluginIconPath: "/assets/peek_plugin_noop/home_icon.png"
            }
        ];
        """

        links = []
        for pluginDetail in pluginDetails:
            if not pluginDetail.cfgModule:
                continue

            links.append(dict(name=pluginDetail.pluginName,
                              title=pluginDetail.homeLinkText,
                              resourcePath="/%s_cfg/" % pluginDetail.pluginName,
                              pluginIconPath=pluginDetail.icon))

        links.sort(key=lambda item: item["title"])

        contents = "// This file is auto generated, the git version is blank and .gitignored\n"
        contents += "export const configLinks = %s;\n" % json.dumps(
            links, sort_keys=True, indent=4, separators=(', ', ': '))

        self._writeFileIfRequired(feAppDir, 'plugin-config-links.ts', contents)

    def _writePluginTitleBarLinks(self, feAppDir: str,
                                  pluginDetails: [PluginDetail]) -> None:
        """
        
        import {TitleBarLink} from "@synerty/peek-plugin-base-js";

        export const titleBarLinks :TitleBarLink = [
            {
                plugin : "peek_plugin_noop",
                text: "Noop",
                left: false,
                resourcePath: "/peek_plugin_noop/home_icon.png",
                badgeCount : null
            }
        ];
        """

        links = []
        for pluginDetail in pluginDetails:
            if not (pluginDetail.appModule and pluginDetail.showInTitleBar):
                continue

            links.append(dict(plugin=pluginDetail.pluginName,
                              text=pluginDetail.titleBarText,
                              left=pluginDetail.titleBarLeft,
                              resourcePath="/%s" % pluginDetail.pluginName,
                              badgeCount=None))

        contents = "// This file is auto generated, the git version is blank and .gitignored\n\n"
        contents += "import { ITitleBarLink } from '@synerty/peek-plugin-base-js'\n\n"
        contents += "export const titleBarLinks: ITitleBarLink[] = %s;\n" % json.dumps(
            links, sort_keys=True, indent=4, separators=(', ', ': '))

        self._writeFileIfRequired(feAppDir, 'plugin-title-bar-links.ts', contents)

    def _writePluginAppRouteLazyLoads(self, feAppDir: str,
                                      pluginDetails: [PluginDetail]) -> None:
        _appRoutesTemplate = dedent("""
            {
                path: '%s',
                loadChildren: "%s/%s"
            }""")

        routes = []
        for pluginDetail in pluginDetails:
            if pluginDetail.appModule:
                routes.append(_appRoutesTemplate
                              % (pluginDetail.pluginName,
                                 pluginDetail.pluginName,
                                 pluginDetail.appModule))

        routeData = "// This file is auto generated, the git version is blank and .gitignored\n"
        routeData += "export const pluginAppRoutes = ["
        routeData += ",".join(routes)
        routeData += "\n];\n"

        self._writeFileIfRequired(feAppDir, 'plugin-app-routes.ts', routeData)

    def _writePluginCfgRouteLazyLoads(self, feAppDir: str,
                                      pluginDetails: [PluginDetail]) -> None:
        _cfgRoutesTemplate = dedent("""
            {
                path: '%s_cfg',
                loadChildren: "%s_cfg/%s"
            }""")

        routes = []
        for pluginDetail in pluginDetails:
            if pluginDetail.cfgModule:
                routes.append(_cfgRoutesTemplate
                              % (pluginDetail.pluginName,
                                 pluginDetail.pluginName,
                                 pluginDetail.cfgModule))

        routeData = "// This file is auto generated, the git version is blank and .gitignored\n"
        routeData += "export const pluginCfgRoutes = ["
        routeData += ",".join(routes)
        routeData += "\n];\n"

        self._writeFileIfRequired(feAppDir, 'plugin-cfg-routes.ts', routeData)

    def _writePluginRootModules(self, feAppDir: str,
                                pluginDetails: [PluginDetail]) -> None:

        # initialise the arrays, and put in the persisted service module
        imports = ['''import {PluginRootServicePersistentLoadModule} 
                        from "./plugin-root-services";''']
        modules = ['PluginRootServicePersistentLoadModule']

        for pluginDetail in pluginDetails:
            for rootModule in pluginDetail.rootModules:
                filePath = self._makeModuleOrServicePath(pluginDetail, rootModule)
                imports.append('import {%s} from "%s/%s";'
                               % (rootModule["class"],
                                  filePath,
                                  rootModule["file"]))
                modules.append(rootModule["class"])

        routeData = "// This file is auto generated, the git version is blank and .gitignored\n"
        routeData += '\n'.join(imports) + '\n'
        routeData += "export const pluginRootModules = [\n\t"
        routeData += ",\n\t".join(modules)
        routeData += "\n];\n"

        self._writeFileIfRequired(feAppDir, 'plugin-root-modules.ts', routeData)

    def _writePluginRootServices(self, feAppDir: str,
                                 pluginDetails: [PluginDetail]) -> None:

        imports = []
        services = []
        persistentServices = []
        for pluginDetail in pluginDetails:
            for rootService in pluginDetail.rootServices:
                filePath = self._makeModuleOrServicePath(pluginDetail, rootService)
                imports.append('import {%s} from "%s/%s";'
                               % (rootService["class"],
                                  filePath,
                                  rootService["file"]))

                if rootService["useClassFile"] and rootService["useClassClass"]:
                    imports.append('import {%s} from "%s/%s";'
                                   % (rootService["useClassClass"],
                                      filePath,
                                      rootService["useClassFile"]))
                    services.append(
                        '{provide:%s, useClass:%s}'
                        % (rootService["class"], rootService["useClassClass"])
                    )

                elif rootService["useExistingClass"]:
                    services.append(
                        '{provide:%s, useExisting:%s}'
                        % (rootService["class"], rootService["useExistingClass"])
                    )

                else:
                    services.append(rootService["class"])

                if rootService["persistent"]:
                    persistentServices.append(rootService["class"])

        routeData = "// This file is auto generated, the git version is blank and .gitignored\n"
        routeData += '\n'.join(imports) + '\n'
        routeData += "export const pluginRootServices = [\n\t"
        routeData += ",\n\t".join(services)
        routeData += "\n];\n"

        routeData += '''
        import {NgModule} from "@angular/core";

        @NgModule({
        
        })
        export class PluginRootServicePersistentLoadModule {
            constructor(%s){
        
            }
        
        }
        ''' % ', '.join(['private _%s:%s' % (s, s) for s in persistentServices])

        self._writeFileIfRequired(feAppDir, 'plugin-root-services.ts', routeData)

    def _writePluginRootComponents(self, feAppDir: str,
                                   pluginDetails: [PluginDetail]) -> None:

        # initialise the arrays, and put in the persistent service module
        selectors = []

        for pluginDetail in pluginDetails:
            for comp in pluginDetail.rootComponents:
                selectors.append('    <%(s)s></%(s)s>' % dict(s=comp["selector"]))

        html = "<div>\n%s\n</div>\n" % '\n'.join(selectors)
        nsXml = "<StackLayout>\n%s\n</StackLayout>\n" % '\n'.join(selectors)

        if self._buildType == BuildTypeEnum.NATIVE_SCRIPT:
            self._writeFileIfRequired(feAppDir, 'plugin-root.component.ns.html', nsXml)
        else:
            self._writeFileIfRequired(feAppDir, 'plugin-root.component.web.html', html)

    def _syncPluginFiles(self, targetDir: str,
                         pluginDetails: [PluginDetail],
                         attrName: str,
                         preSyncCallback: Optional[Callable[[], None]] = None,
                         postSyncCallback: Optional[Callable[[], None]] = None,
                         keepCompiledFilePatterns: Optional[Dict[str, List[str]]] = None,
                         excludeFilesRegex=(),
                         isCfgDir=False) -> None:
        cfgPostfix = "_cfg"

        if not os.path.exists(targetDir):
            os.mkdir(targetDir)  # The parent must exist

        # Make a note of the existing items
        currentItems = set()
        createdItems = set()
        for item in os.listdir(targetDir):
            if not item.startswith("peek_plugin_") or item.startswith("peek_core_"):
                continue

            if isCfgDir:
                if item.endswith(cfgPostfix):
                    currentItems.add(item)
            else:
                if not item.endswith(cfgPostfix):
                    currentItems.add(item)

        for pluginDetail in pluginDetails:
            frontendDir = getattr(pluginDetail, attrName, None)
            if not frontendDir:
                continue

            srcDir = os.path.join(pluginDetail.pluginRootDir, frontendDir)
            if not os.path.exists(srcDir):
                logger.warning("%s FE dir %s doesn't exist",
                               pluginDetail.pluginName, frontendDir)
                continue

            if isCfgDir:
                createdItems.add(pluginDetail.pluginName + cfgPostfix)
                linkPath = os.path.join(targetDir, pluginDetail.pluginName + cfgPostfix)
            else:
                createdItems.add(pluginDetail.pluginName)
                linkPath = os.path.join(targetDir, pluginDetail.pluginName)

            self.fileSync.addSyncMapping(srcDir, linkPath,
                                         keepCompiledFilePatterns=keepCompiledFilePatterns,
                                         preSyncCallback=preSyncCallback,
                                         postSyncCallback=postSyncCallback,
                                         excludeFilesRegex=excludeFilesRegex)

        # Delete the items that we didn't create
        for item in currentItems - createdItems:
            path = os.path.join(targetDir, item)
            if os.path.islink(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)

    def _updatePackageJson(self, targetJson: str,
                           pluginDetails: [PluginDetail]) -> None:

        serviceName = "@peek"

        # Remove all the old symlinks

        with open(targetJson, 'r') as f:
            jsonData = json.load(f)

        dependencies = jsonData["dependencies"]
        for key in list(dependencies):
            if key.startswith(serviceName):
                del dependencies[key]

        for pluginDetail in pluginDetails:
            if not pluginDetail.moduleDir:
                continue

            moduleDir = os.path.join(pluginDetail.pluginRootDir,
                                     pluginDetail.moduleDir)

            name = "%s/%s" % (serviceName, pluginDetail.pluginName)
            dependencies[name] = "file:///" + moduleDir.replace("\\", '/')

        contents = json.dumps(jsonData, sort_keys=True, indent=2,
                              separators=(',', ': '))

        self._writeFileIfRequired(os.path.dirname(targetJson),
                                  os.path.basename(targetJson),
                                  contents)



import logging
from datetime import datetime

import os
from typing import List

import pytz

from peek_platform.build_frontend.FrontendBuilderABC import FrontendBuilderABC, BuildTypeEnum
from peek_platform.build_common.BuilderOsCmd import runNgBuild
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class WebBuilder(FrontendBuilderABC):

    def __init__(self, frontendProjectDir: str, platformService: str,
                 jsonCfg, loadedPlugins: List):
        FrontendBuilderABC.__init__(self, frontendProjectDir, platformService,
                                    self._buildType(platformService),
                                    jsonCfg, loadedPlugins)

        self.isMobile = "mobile" in platformService
        self.isDesktop = "desktop" in platformService
        self.isAdmin = "admin" in platformService

    @staticmethod
    def _buildType(platformService: str):
        if "mobile" in platformService: return BuildTypeEnum.WEB_MOBILE
        if "desktop" in platformService: return BuildTypeEnum.WEB_DESKTOP
        if "admin" in platformService: return BuildTypeEnum.WEB_ADMIN

        raise NotImplementedError("Unknown build type")

    @deferToThreadWrapWithLogger(logger, checkMainThread=False)
    def build(self) -> None:
        if not self._jsonCfg.feWebBuildPrepareEnabled:
            logger.info("%s SKIPPING, Web build prepare is disabled in config",
                        self._platformService)
            return

        excludeRegexp = (
            r'.*[.]ns[.]ts$',
            r'.*[.]ns[.]html$',
            r'.*[.]ns[.]scss$',
            r'.*[.]ns[.].*[.]scss$',
            r'.*__pycache__.*',
            r'.*[.]py$'
        )

        if self.isMobile:
            excludeRegexp += (
                r'.*[.]dweb[.]ts$',
                r'.*[.]dweb[.]html$',
                r'.*[.]dweb[.]scss',
            )

        elif self.isDesktop:
            excludeRegexp += (
                r'.*[.]mweb[.]ts$',
                r'.*[.]mweb[.]html$',
                r'.*[.]mweb[.]scss',
            )

        elif self.isAdmin:
            pass

        else:
            raise NotImplementedError("This is neither admin, mobile or desktop web")

        self._dirSyncMap = list()

        feBuildDir = os.path.join(self._frontendProjectDir, 'build-web')
        feSrcAppDir = os.path.join(self._frontendProjectDir, 'src', 'app')

        feBuildSrcDir = os.path.join(feBuildDir, 'src')
        feBuildAssetsDir = os.path.join(feBuildDir, 'src', 'assets')
        feNodeModDir = os.path.join(feBuildDir, 'node_modules')

        feModuleDirs = [
            (os.path.join(feBuildSrcDir, '@peek'), "moduleDir"),
        ]

        pluginDetails = self._loadPluginConfigs()

        # --------------------
        # Check if node_modules exists

        if not os.path.exists(os.path.join(feBuildDir, 'node_modules')):
            raise NotADirectoryError(
                "%s node_modules doesn't exist, ensure you've run "
                "`npm install` in dir %s",
                self._platformService, feBuildDir)

        # --------------------
        # Prepare the common frontend application

        self.fileSync.addSyncMapping(feSrcAppDir, os.path.join(feBuildSrcDir, 'app'),
                                     excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the home and title bar configuration for the plugins
        self._writePluginHomeLinks(feBuildSrcDir, pluginDetails)
        self._writePluginTitleBarLinks(feBuildSrcDir, pluginDetails)
        self._writePluginConfigLinks(feBuildSrcDir, pluginDetails)

        # --------------------
        # Prepare the plugin lazy loaded part of the application
        self._writePluginAppRouteLazyLoads(feBuildSrcDir, pluginDetails)
        self._syncPluginFiles(feBuildSrcDir, pluginDetails, "appDir",
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the plugin lazy loaded part of the application
        self._writePluginCfgRouteLazyLoads(feBuildSrcDir, pluginDetails)
        self._syncPluginFiles(feBuildSrcDir, pluginDetails, "cfgDir",
                              isCfgDir=True,
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the plugin assets
        self._syncPluginFiles(feBuildAssetsDir, pluginDetails, "assetDir",
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the shared / global parts of the plugins

        self._writePluginRootModules(feBuildSrcDir, pluginDetails)
        self._writePluginRootServices(feBuildSrcDir, pluginDetails)
        self._writePluginRootComponents(feBuildSrcDir, pluginDetails)

        for feModDir, jsonAttr, in feModuleDirs:
            # Link the shared code, this allows plugins
            # * to import code from each other.
            # * provide global services.
            self._syncPluginFiles(feModDir, pluginDetails, jsonAttr,
                                  excludeFilesRegex=excludeRegexp)

        # Lastly, Allow the clients to override any frontend files they wish.
        # Src Directory
        self.fileSync.addSyncMapping(self._jsonCfg.feFrontendSrcOverlayDir,
                                     feBuildSrcDir,
                                     parentMustExist=True,
                                     deleteExtraDstFiles=False,
                                     excludeFilesRegex=excludeRegexp)

        # node_modules Directory
        self.fileSync.addSyncMapping(self._jsonCfg.feFrontendNodeModuleOverlayDir,
                                     feNodeModDir,
                                     parentMustExist=True,
                                     deleteExtraDstFiles=False,
                                     excludeFilesRegex=excludeRegexp)

        self.fileSync.syncFiles()

        if self._jsonCfg.feSyncFilesForDebugEnabled:
            logger.info("%s starting frontend development file sync",
                        self._platformService)
            self.fileSync.startFileSyncWatcher()

        if self._jsonCfg.feWebBuildEnabled:
            logger.info("%s starting frontend web build", self._platformService)
            self._compileFrontend(feBuildDir)

    def _syncFileHook(self, fileName: str, contents: bytes) -> bytes:
        # replace imports that end with .dweb or .mweb to the appropriate
        # value
        # Otherwise just .web should be used if no replacing is required.

        if self.isMobile:
            contents = contents.replace(b'.dweb";', b'.mweb";')

        elif self.isDesktop:
            contents = contents.replace(b'.mweb";', b'.dweb";')

        elif self.isAdmin:
            pass

        else:
            raise NotImplementedError("This is neither mobile or desktop web")

        if b'@Component' in contents:
            return self._patchComponent(fileName, contents)

        return contents

    def _patchComponent(self, fileName: str, contents: bytes) -> bytes:
        """ Patch Component

        Apply patches to the WEB file to convert it to the NativeScript version

        :param fileName: The name of the file
        :param contents: The original contents of the file
        :return: The new contents of the file
        """
        inComponentHeader = False

        newContents = b''
        for line in contents.splitlines(True):
            if line.startswith(b"@Component"):
                inComponentHeader = True

            elif line.startswith(b"export"):
                inComponentHeader = False

            elif inComponentHeader:

                if self.isDesktop:
                    line = (line
                        .replace(b'.mweb.html', b'.dweb.html')
                        .replace(b'.mweb.css', b'.dweb.css')
                        .replace(b'.mweb.scss', b'.dweb.scss')
                        )

                if self.isMobile:
                    line = (line
                        .replace(b'.dweb.html', b'.mweb.html')
                        .replace(b'.dweb.css', b'.mweb.css')
                        .replace(b'.dweb.scss', b'.mweb.scss')
                        )

            newContents += line

        return newContents

    def _compileFrontend(self, feBuildDir: str) -> None:
        """ Compile the frontend

        this runs `ng build`

        We need to use a pty otherwise webpack doesn't run.

        """
        startDate = datetime.now(pytz.utc)
        hashFileName = os.path.join(feBuildDir, ".lastHash")

        if not self._recompileRequiredCheck(feBuildDir, hashFileName):
            logger.info("%s Frontend has not changed, recompile not required.",
                        self._platformService)
            return

        logger.info("%s Rebuilding frontend distribution", self._platformService)

        try:
            runNgBuild(feBuildDir)

        except Exception as e:
            if os.path.exists(hashFileName):
                os.remove(hashFileName)

            # Update the detail of the exception and raise it
            e.message = "%s angular frontend failed to build." % self._platformService
            raise

        logger.info("%s frontend rebuild completed in %s",
                    self._platformService, datetime.now(pytz.utc) - startDate)



import logging
import typing

import celery
from celery import signals as celery_signals
from kombu import serialization
from peek_platform.file_config.PeekFileConfigWorkerMixin import PeekFileConfigWorkerMixin
from vortex.DeferUtil import noMainThread
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


def vortexDumps(arg: typing.Tuple) -> str:
    noMainThread()
    # startTime = datetime.now(pytz.utc)
    try:
        return Payload(tuples=[arg])._toJson()
    except Exception as e:
        logger.exception(e)
        raise
    # finally:
    #     logger.debug("vortexDumps took %s", (datetime.now(pytz.utc) - startTime))


def vortexLoads(jsonStr: str) -> typing.Tuple:
    noMainThread()
    # startTime = datetime.now(pytz.utc)
    try:
        return Payload()._fromJson(jsonStr).tuples[0]
    except Exception as e:
        logger.exception(e)
        raise
    # finally:
    #     logger.debug("vortexLoads took %s", (datetime.now(pytz.utc) - startTime))


serialization.register(
    'vortex', vortexDumps, vortexLoads,
    content_type='application/x-vortex',
    content_encoding='utf-8',
)


# -----------------------------------------------------------------------------

class BackendMixin:

    def exception_to_python(self, exc):
        """Convert serialized exception to Python exception."""
        import sys
        from kombu.utils.encoding import from_utf8
        from celery.utils.serialization import (create_exception_cls,
                                                get_pickled_exception)

        EXCEPTION_ABLE_CODECS = frozenset({'pickle'})

        if not exc:
            return exc

        if not isinstance(exc, BaseException):
            exc_module = exc.get('exc_module')
            if exc_module is None:
                cls = create_exception_cls(
                    from_utf8(exc['exc_type']), __name__)
            else:
                exc_module = from_utf8(exc_module)
                exc_type = from_utf8(exc['exc_type'])
                try:
                    cls = getattr(sys.modules[exc_module], exc_type)
                except KeyError:
                    cls = create_exception_cls(exc_type,
                                               celery.exceptions.__name__)
            exc_msg = exc['exc_message']
            args = exc_msg if isinstance(exc_msg, tuple) else [exc_msg]

            ### BEGIN CODE ADDED TO PATCH METHOD
            try:
                exc = cls(*args)

            except TypeError:
                exc = Exception("%s\n%s" % (cls, args))
            ### END CODE ADDED TO PATCH METHOD

            if self.serializer in EXCEPTION_ABLE_CODECS:
                exc = get_pickled_exception(exc)

        return exc


from celery.backends.base import Backend

Backend.exception_to_python = BackendMixin.exception_to_python


# -----------------------------------------------------------------------------

class ResultConsumerMixin:

    def cancel_for(self, task_id):
        import redis

        if not self._pubsub:
            return

        key = self._get_key_for_task(task_id)
        self.subscribed_to.discard(key)

        try:
            self._pubsub.unsubscribe(key)

        except redis.exceptions.ConnectionError:
            # redis.exceptions.ConnectionError: Error 32 while writing to socket. Broken pipe.
            logger.debug("Cancel_for failed due to redis error")

        except RuntimeError:
            # builtins.RuntimeError: dictionary changed size during iteration
            logger.debug("Cancel_for failed due to redis error")

        except AttributeError:
            # builtins.AttributeError: 'NoneType' object has no attribute 'sendall'
            logger.debug("Cancel_for failed due to redis error")


from celery.backends.redis import ResultConsumer

ResultConsumer.cancel_for = ResultConsumerMixin.cancel_for


# -----------------------------------------------------------------------------

def configureCeleryApp(app, workerConfig: PeekFileConfigWorkerMixin,
                       forCaller: bool = False):
    # Optional configuration, see the application user guide.
    app.conf.update(
        # On peek_server, the thread limit is set to 10, these should be configurable.
        # broker_pool_limit=20,

        # Set the broker and backend URLs
        broker_url=workerConfig.celeryBrokerUrl,
        result_backend=workerConfig.celeryResultUrl,

        # Leave the logging to us
        worker_hijack_root_logger=False,

        # The time results will stay in redis before expiring.
        # I believe they are cleared when the results are obtained
        # from txcelery._DeferredTask
        # I assume the timer only starts once the task has finished.
        result_expires=60,

        task_serializer='vortex',
        # accept_content=['vortex'],  # Ignore other content
        accept_content=['pickle', 'json', 'msgpack', 'yaml', 'vortex'],
        result_serializer='vortex',
        enable_utc=True,

        # Default time in seconds before a retry of the task should be executed.
        # 3 minutes by default.
        default_retry_delay=2,

        # The maximum number of times to retry a task
        max_retries=5
    )

    # Configure these only for the worker, this keeps the servers json clean.
    if not forCaller:
        # Optional configuration, see the application user guide.
        app.conf.update(
            # The number of tasks each worker will prefetch.
            worker_prefetch_multiplier=workerConfig.celeryTaskPrefetch,

            # The number of workers to have at one time
            worker_concurrency=workerConfig.celeryWorkerCount,
        )

        if workerConfig.celeryReplaceWorkerAfterTaskCount:
            app.conf.update(
                # The number of tasks a worker will process before it's replaced
                worker_max_tasks_per_child=workerConfig.celeryReplaceWorkerAfterTaskCount,
            )

        if workerConfig.celeryReplaceWorkerAfterMemUsage:
            app.conf.update(
                # If a worker uses more than this amount of memory, it will be replaced
                # after the task completes.
                worker_max_memory_per_child=workerConfig.celeryReplaceWorkerAfterMemUsage,
            )


from peek_platform.file_config.PeekFileConfigABC import PeekFileConfigABC
from peek_platform.file_config.PeekFileConfigPlatformMixin import \
    PeekFileConfigPlatformMixin


class _WorkerTaskConfigMixin(PeekFileConfigABC,
                             PeekFileConfigPlatformMixin):
    pass


@celery_signals.after_setup_logger.connect
def configureCeleryLogging(*args, **kwargs):
    from peek_plugin_base.PeekVortexUtil import peekWorkerName

    # Fix the loading problems windows has
    # from peek_platform.util.LogUtil import setupPeekLogger
    # setupPeekLogger(peekWorkerName)

    from peek_platform import PeekPlatformConfig
    PeekPlatformConfig.componentName = peekWorkerName
    config = _WorkerTaskConfigMixin()

    # Set default logging level
    logging.root.setLevel(config.loggingLevel)

    if config.loggingLevel != "DEBUG":
        for name in ("celery.worker.strategy", "celery.app.trace", "celery.worker.job"):
            logging.getLogger(name).setLevel(logging.WARNING)
